#include <mpi.h>
#include <stdio.h>
#include <omp.h>

double do_some_integratin(long long nsize,int myrank,int nprocs)
{

  long long i, start_int, end_int;
  int       iam, np;
  double    mypi, h, x;
  double    sum = 0.;
  
  /* Calculate the interval size */
  h = (double)1.0/(double)(nsize*nprocs);
   
  /* Calculate the interval section */
  start_int = nsize*myrank + 1;
  end_int = start_int + nsize;
  
  /* Begin the OpenMP section */
#pragma omp parallel private(iam,x,i,np)
  {

#pragma omp barrier
    iam = omp_get_thread_num() ;
    np=omp_get_num_threads() ;
  
    /* Output the # of tasks and threads */
    //printf("Thread:%5d rank:%5d thread:%5d processes:%5d\n",iam,myrank,np,nprocs); 
    //printf("Thread %5d of %5d in MPI task %5d of %5d\n",iam,np,myrank,nprocs); 
  
#pragma omp barrier
  
    /* integrate over the appropriate interval over np threads */
#pragma omp for schedule(static),reduction(+:sum)
    for(i=start_int;i<=end_int;i++)
    {
      x = h * ((double)i - 0.5);
      sum = sum + (4./(1. + x*x));
    }
  } 
  
  mypi = h * sum;
  return mypi;
}     
     
int main( argc, argv )
int argc;
char **argv;
{
  
  int        nprocs, myrank;
  long long totaln=50000000000;
  //long long  nsize=201000000;
  double     mypi=0, pi=0;
  
  double t1, t2, t3, t4, t5;
  t1 = omp_get_wtime();

  MPI_Init( &argc, &argv );
  MPI_Comm_rank( MPI_COMM_WORLD, &myrank );
  MPI_Comm_size( MPI_COMM_WORLD, &nprocs );
  
  long long nsize=totaln/nprocs;
  //printf("size = %lld in rank %d \n",nsize,myrank);
  
  MPI_Barrier(MPI_COMM_WORLD);
  if ( myrank == 0 ) printf("\n");
  
  t2 = omp_get_wtime();
  /*  Compute pi = integral(4/(1+x2) on the interval (0,1) */
  mypi = do_some_integratin(nsize,myrank,nprocs);
  t3 = omp_get_wtime();
  
  /*  Collect all the partial sums */
  MPI_Reduce(&mypi,&pi,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
  t4 = omp_get_wtime();
  
  MPI_Barrier(MPI_COMM_WORLD);
  
  if (myrank == 0) printf ("The value of pi is: %11.8f\n",pi);
  
  MPI_Finalize();
  t5 = omp_get_wtime();
  
  printf("\n --------------- My rank is: %d/%d ---------------",myrank,nprocs);
  //printf("\n pre-MPI time     = %9.5f seconds.", t2-t1 );
  printf("\n integration time = %9.5f seconds. \n", t3-t2 );
  //printf("\n MPI_reduce time  = %9.5f seconds.", t4-t3 );
  //printf("\n post MPI time    = %9.5f seconds.", t5-t4 );
  //printf("\n total time       = %9.5f seconds.", t5-t1 );

  return 0;
}
