// ****************************************************************************
// offload_transfer demo
// compile with: icc -openmp
//
// djames (at) tacc.utexas.edu
// 28 Sep 2013
//
//    09 Jul 2014   Fixed alloc/free settings to originally intended values
//
// ****************************************************************************

#include <stdio.h>
#include <omp.h>
#include <malloc.h>

// -------------------------------------------------------------------

int main( void ) {

  int i;
  int N = 5000000;
  double *a, *b;
  int tag1 = 1, tag2 = 2, tag3 = 3;
    // Intel source code says assign unique values; can't find this in 
    // official documentation anywhere, and it doesn't seem to matter here.

  a = ( double* ) memalign( 64, N*sizeof(double) );  // allocate alligned memory on host
  b = ( double* ) memalign( 64, N*sizeof(double) );

// allocate memory on mic
  #pragma offload_transfer target(mic:0)           \
   nocopy( a : length(N) alloc_if(1) free_if(0) ) \
   nocopy( b : length(N) alloc_if(1) free_if(0) )  \
   signal( &tag1 )

  for ( i=0; i<N; i++ )  {  // asign value to a on host
    a[i] = (double)(i);
  }

// after tag1 is finished, copy a from host to mic, calculate b on mic, copy b from mic to host
  #pragma offload target(mic:0) wait( &tag1 )      \
   in(    a : length(N) alloc_if(0) free_if(0) ) \
   out(   b : length(N) alloc_if(0) free_if(0) )  \
   signal( &tag2 )
    #pragma omp parallel for
      for ( i=0; i<N; i++ )  {
        b[i] = 2.0 * a[i];
      }

// after tag2 is finished, copy b from mic to host, deallocate a and b on mic
  #pragma offload_transfer target(mic:0) wait( &tag2 ) \
   nocopy( a : length(N) alloc_if(0) free_if(1) )    \
   out( b : length(N) alloc_if(0) free_if(1) )    \
   signal( &tag3 )


// wait for tag3 finished
  #pragma offload_wait target(mic:0) wait( &tag3 )
  {
    printf( "\n\t last a val  = %f",     a[N-1]);
    printf( "\n\t last b val  = %f\n\n", b[N-1]);
  }

  free(a);  // decallocate a and b on host
  free(b);

  return 0;

}
